"""
Copyright (c) 2022, Oracle and/or its affiliates.
Copyright (c) 2020, Vitor Avancini

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
"""

import pytest

import datetime

from dbt.tests.util import run_dbt, relation_from_name


# seeds/my_seed.csv
seed_csv = """
user_id,user_name,birth_date,income,last_login_date
1,Easton,1981-05-20,40000,2022-04-25T08:57:59
2,Lillian,1978-09-03,54000,2022-04-25T08:58:59
3,Jeremiah,1982-03-11,10000,2022-04-25T09:57:59
4,Nolan,1976-05-06,900000,2022-04-25T09:58:59
""".lstrip()

# models/foo.sql
my_model_foo_sql = """
{{config(materialized='table', 
         alias="foo")}}
SELECT * FROM {{ ref('seed') }}
"""

# models/foo2.sql
my_model_foo2_sql = """
{{config(materialized='table', 
         alias="Foo")}}
SELECT * FROM {{ ref('seed') }}
"""


class TestQuotedLowerCaseTableName:

    @pytest.fixture(scope="class")
    def seeds(self):
        return {
            "seed.csv": seed_csv,
        }

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "my_model_foo.sql": my_model_foo_sql,
        }

    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "quoting": {
                "database": False,
                "schema": False,
                "identifier": True
            },
        }

    def test_run_dbt(self, project):
        """
            - run seed
            - run model

        """
        results = run_dbt(['seed'])
        assert len(results) == 1

        results = run_dbt(['run'])
        assert len(results) == 1

        results = run_dbt(['run'])
        assert len(results) == 1

        sql = 'SELECT COUNT(*) from "foo"'
        result = project.run_sql(sql, fetch="all")
        assert result == [(4,)]


class TestQuotedTitleCaseTableName:

    @pytest.fixture(scope="class")
    def seeds(self):
        return {
            "seed.csv": seed_csv,
        }

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "my_model_foo.sql": my_model_foo2_sql,
        }

    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "quoting": {
                "database": False,
                "schema": False,
                "identifier": True
            },
        }

    def test_run_dbt(self, project):
        """
            - run seed
            - run model

        """
        results = run_dbt(['seed'])
        assert len(results) == 1

        results = run_dbt(['run'])
        assert len(results) == 1

        results = run_dbt(['run'])
        assert len(results) == 1

        sql = 'SELECT COUNT(*) from "Foo"'
        result = project.run_sql(sql, fetch="all")
        assert result == [(4,)]


class TestUnQuotedLowerCaseTableName:

    @pytest.fixture(scope="class")
    def seeds(self):
        return {
            "seed.csv": seed_csv,
        }

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "my_model_foo.sql": my_model_foo_sql,
        }

    def test_run_dbt(self, project):
        """
            - run seed
            - run model

        """
        results = run_dbt(['seed'])
        assert len(results) == 1

        results = run_dbt(['run'])
        assert len(results) == 1

        results = run_dbt(['run'])
        assert len(results) == 1

        sql = 'SELECT COUNT(*) from FOO'
        result = project.run_sql(sql, fetch="all")
        assert result == [(4,)]


class TestUnQuotedTitleCaseTableName:

    @pytest.fixture(scope="class")
    def seeds(self):
        return {
            "seed.csv": seed_csv,
        }

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "my_model_foo.sql": my_model_foo2_sql,
        }

    def test_run_dbt(self, project):
        """
            - run seed
            - run model

        """
        results = run_dbt(['seed'])
        assert len(results) == 1

        results = run_dbt(['run'])
        assert len(results) == 1

        results = run_dbt(['run'])
        assert len(results) == 1

        sql = 'SELECT COUNT(*) from FOO'
        result = project.run_sql(sql, fetch="all")
        assert result == [(4,)]





"""
Copyright (c) 2022, Oracle and/or its affiliates.
Copyright (c) 2020, Vitor Avancini

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
"""
from typing import Optional

from dataclasses import dataclass, field

from dbt.adapters.base.relation import BaseRelation
from dbt.adapters.events.logging import AdapterLogger
from dbt.adapters.relation_configs import (
    RelationConfigBase,
    RelationConfigChangeAction,
    RelationResults,
)
from dbt.context.providers import RuntimeConfigObject
from dbt.contracts.graph.nodes import ModelNode
from dbt.adapters.base import RelationType
from dbt.exceptions import DbtRuntimeError

from dbt.adapters.oracle.relation_configs import (
    OracleMaterializedViewConfig,
    OracleRefreshMethodConfigChange,
    OracleMaterializedViewConfigChangeset,
    OracleRefreshModeConfigChange,
    OracleBuildModeConfigChange,
    OracleQueryRewriteConfigChange,
    OracleQueryConfigChange,
    OracleQuotePolicy,
    OracleIncludePolicy)



logger = AdapterLogger("oracle")


@dataclass(frozen=True, eq=False, repr=False)
class OracleRelation(BaseRelation):
    quote_policy: OracleQuotePolicy = field(default_factory=lambda: OracleQuotePolicy())
    include_policy: OracleIncludePolicy = field(default_factory=lambda: OracleIncludePolicy())
    relation_configs = {
        RelationType.MaterializedView.value: OracleMaterializedViewConfig
    }

    @staticmethod
    def add_ephemeral_prefix(name):
        return f'dbt__cte__{name}__'

    @classmethod
    def from_runtime_config(cls, runtime_config: RuntimeConfigObject) -> RelationConfigBase:
        model_node: ModelNode = runtime_config.model
        relation_type: str = model_node.config.materialized

        if relation_config := cls.relation_configs.get(relation_type):
            return relation_config.from_model_node(model_node)

        raise DbtRuntimeError(
            f"from_runtime_config() is not supported for the provided relation type: {relation_type}"
        )

    @classmethod
    def materialized_view_config_changeset(
            cls, relation_results: RelationResults,
            runtime_config: RuntimeConfigObject) -> Optional[OracleMaterializedViewConfigChangeset]:
        config_change_collection = OracleMaterializedViewConfigChangeset()

        existing_materialized_view = OracleMaterializedViewConfig.from_relation_results(
            relation_results
        )
        new_materialized_view = OracleMaterializedViewConfig.from_model_node(
            runtime_config.model
        )

        assert isinstance(existing_materialized_view, OracleMaterializedViewConfig)
        assert isinstance(new_materialized_view, OracleMaterializedViewConfig)

        if new_materialized_view.refresh_method.upper() != existing_materialized_view.refresh_method.upper():
            config_change_collection.refresh_method = OracleRefreshMethodConfigChange(
                action=RelationConfigChangeAction.alter,
                context=new_materialized_view.refresh_method
            )

        if new_materialized_view.refresh_mode.upper() != existing_materialized_view.refresh_mode.upper():
            config_change_collection.refresh_mode = OracleRefreshModeConfigChange(
                action=RelationConfigChangeAction.alter,
                context=new_materialized_view.refresh_mode
            )

        if new_materialized_view.build_mode.upper() != existing_materialized_view.build_mode.upper():
            config_change_collection.build_mode = OracleBuildModeConfigChange(
                action=RelationConfigChangeAction.alter,
                context=new_materialized_view.build_mode
            )

        if new_materialized_view.query_rewrite.upper() != existing_materialized_view.query_rewrite.upper():
            config_change_collection.query_rewrite = OracleQueryRewriteConfigChange(
                action=RelationConfigChangeAction.alter,
                context=new_materialized_view.query_rewrite
            )

        if new_materialized_view.query.upper() != existing_materialized_view.query.upper():
            config_change_collection.query = OracleQueryConfigChange(
                action=RelationConfigChangeAction.create,
                context=new_materialized_view.query
            )

        logger.debug(f"Config change collection {config_change_collection}")

        if config_change_collection.has_changes:

            if config_change_collection.refresh_mode is None:
                config_change_collection.refresh_mode = OracleRefreshModeConfigChange(
                    action=RelationConfigChangeAction.alter,
                    context=existing_materialized_view.refresh_mode)

            if config_change_collection.query_rewrite is None:
                config_change_collection.query_rewrite = OracleQueryRewriteConfigChange(
                    action=RelationConfigChangeAction.alter,
                    context=existing_materialized_view.query_rewrite)

            if config_change_collection.refresh_method is None:
                config_change_collection.refresh_method = OracleRefreshMethodConfigChange(
                    action=RelationConfigChangeAction.alter,
                    context=existing_materialized_view.refresh_method)

            if config_change_collection.build_mode is None:
                config_change_collection.build_mode = OracleBuildModeConfigChange(
                    action=RelationConfigChangeAction.alter,
                    context=existing_materialized_view.build_mode)

            if config_change_collection.query is None:
                config_change_collection.query = OracleQueryConfigChange(
                    action=RelationConfigChangeAction.create,
                    context=existing_materialized_view.query)

            return config_change_collection

        return None



import pytest

from dbt.tests.adapter.basic.test_docs_generate import (BaseDocsGenReferences,
                                                        ref_models__schema_yml,
                                                        ref_models__ephemeral_copy_sql,
                                                        ref_models__docs_md)

from dbt.tests.adapter.basic.expected_catalog import no_stats

ref_sources__schema_yml = """
version: 2
sources:
  - name: my_source
    description: "{{ doc('source_info') }}"
    loader: a_loader
    schema: "{{ var('test_schema') }}"
    tables:
      - name: my_table
        description: "{{ doc('table_info') }}"
        identifier: seed
        quoting:
          identifier: true
        columns:
          - name: id
            description: "{{ doc('column_info') }}"
"""

ref_models__ephemeral_summary_sql = """
{{
  config(
    materialized = "table"
  )
}}

select "first_name", count(*) as ct from {{ref('ephemeral_copy')}}
group by "first_name"
order by "first_name" asc

"""

ref_models__view_summary_sql = """
{{
  config(
    materialized = "view"
  )
}}

select "first_name", ct from {{ref('ephemeral_summary')}}
order by ct asc

"""


class TextType:

    def __eq__(self, other):
        return other.upper().startswith('VARCHAR')


def expected_references_catalog(
    project,
    role,
    id_type,
    text_type,
    time_type,
    view_type,
    table_type,
    model_stats,
    bigint_type=None,
    seed_stats=None,
    case=None,
    case_columns=False,
    view_summary_stats=None,
):
    if case is None:

        def case(x):
            return x

    col_case = case if case_columns else lambda x: x

    if seed_stats is None:
        seed_stats = model_stats

    if view_summary_stats is None:
        view_summary_stats = model_stats

    model_database = project.database.upper()
    my_schema_name = project.test_schema.upper()

    summary_columns = {
        "first_name": {
            "name": "first_name",
            "index": 1,
            "type": TextType() ,
            "comment": None,
        },
        "CT": {
            "name": "CT",
            "index": 2,
            "type": bigint_type,
            "comment": None,
        },
    }

    seed_columns = {
        "id": {
            "name": col_case("id"),
            "index": 1,
            "type": id_type,
            "comment": None,
        },
        "first_name": {
            "name": col_case("first_name"),
            "index": 2,
            "type": TextType(),
            "comment": None,
        },
        "email": {
            "name": col_case("email"),
            "index": 3,
            "type": TextType(),
            "comment": None,
        },
        "ip_address": {
            "name": col_case("ip_address"),
            "index": 4,
            "type": TextType(),
            "comment": None,
        },
        "updated_at": {
            "name": col_case("updated_at"),
            "index": 5,
            "type": time_type,
            "comment": None,
        },
    }
    return {
        "nodes": {
            "seed.test.seed": {
                "unique_id": "seed.test.seed",
                "metadata": {
                    "schema": my_schema_name,
                    "database": project.database.upper(),
                    "name": case("seed"),
                    "type": table_type,
                    "comment": None,
                    "owner": role,
                },
                "stats": seed_stats,
                "columns": seed_columns,
            },
            "model.test.ephemeral_summary": {
                "unique_id": "model.test.ephemeral_summary",
                "metadata": {
                    "schema": my_schema_name,
                    "database": model_database,
                    "name": case("ephemeral_summary"),
                    "type": table_type,
                    "comment": None,
                    "owner": role,
                },
                "stats": model_stats,
                "columns": summary_columns,
            },
            "model.test.view_summary": {
                "unique_id": "model.test.view_summary",
                "metadata": {
                    "schema": my_schema_name,
                    "database": model_database,
                    "name": case("view_summary"),
                    "type": view_type,
                    "comment": None,
                    "owner": role,
                },
                "stats": view_summary_stats,
                "columns": summary_columns,
            },
        },
        "sources": {
            "source.test.my_source.my_table": {
                "unique_id": "source.test.my_source.my_table",
                "metadata": {
                    "schema": my_schema_name,
                    "database": project.database.upper(),
                    "name": case("seed"),
                    "type": table_type,
                    "comment": None,
                    "owner": role,
                },
                "stats": seed_stats,
                "columns": seed_columns,
            },
        },
    }


class TestOracleDocsGenReferences(BaseDocsGenReferences):

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "schema.yml": ref_models__schema_yml,
            "sources.yml": ref_sources__schema_yml,
            "view_summary.sql": ref_models__view_summary_sql,
            "ephemeral_summary.sql": ref_models__ephemeral_summary_sql,
            "ephemeral_copy.sql": ref_models__ephemeral_copy_sql,
            "docs.md": ref_models__docs_md,
        }

    @pytest.fixture(scope="class")
    def project_config_update(self, unique_schema):
        alternate_schema = unique_schema
        return {
            "asset-paths": ["assets", "invalid-asset-paths"],
            "vars": {
                "test_schema": unique_schema,
                "alternate_schema": alternate_schema,
            },
            "seeds": {
                "quote_columns": True,
            },
            "quoting": {"identifier": True},
        }

    @pytest.fixture(scope="class")
    def expected_catalog(self, project, profile_user):
        return expected_references_catalog(
            project,
            role=profile_user.upper(),
            id_type="NUMBER",
            text_type="VARCHAR2(16)",
            time_type="TIMESTAMP(6)",
            bigint_type="NUMBER",
            view_type="VIEW",
            table_type="BASE TABLE",
            model_stats=no_stats(),
            case_columns=False
        )



"""
Copyright (c) 2023, Oracle and/or its affiliates.
Copyright (c) 2020, Vitor Avancini

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
"""
from typing import List, Optional, Tuple, Any, Dict, Union
from contextlib import contextmanager
from dataclasses import dataclass, field
import json
import enum
import time
import uuid
import platform

import dbt_common.exceptions
from dbt.adapters.contracts.connection import AdapterResponse, Credentials
from dbt.adapters.exceptions.connection import FailedToConnectError
from dbt.adapters.sql import SQLConnectionManager
from dbt.adapters.events.types import ConnectionUsed, SQLQuery, SQLCommit, SQLQueryStatus
from dbt.adapters.events.logging import AdapterLogger

from dbt_common.events.functions import fire_event
from dbt_common.events.contextvars import get_node_info
from dbt_common.utils import cast_to_str

from dbt.version import __version__ as dbt_version
from dbt.adapters.oracle.connection_helper import oracledb, SQLNET_ORA_CONFIG

logger = AdapterLogger("oracle")


DATATYPES = {
    "DB_TYPE_BFILE": "BFILE",
    "DB_TYPE_BINARY_DOUBLE": "BINARY_DOUBLE",
    "DB_TYPE_BINARY_FLOAT": "BINARY_FLOAT",
    "DB_TYPE_BINARY_INTEGER": "BINARY_INTEGER",
    "DB_TYPE_BLOB": "BLOB",
    "DB_TYPE_BOOLEAN": "BOOLEAN",
    "DB_TYPE_CHAR": "CHAR",
    "DB_TYPE_CLOB": "CLOB",
    "DB_TYPE_DATE": "DATE",
    "DB_TYPE_INTERVAL_DS": "INTERVAL DAY TO SECOND",
    "DB_TYPE_INTERVAL_YM": "INTERVAL YEAR TO MONTH",
    "DB_TYPE_JSON": "JSON",
    "DB_TYPE_LONG": "LONG",
    "DB_TYPE_LONG_NVARCHAR": "LONG NVARCHAR",
    "DB_TYPE_LONG_RAW": "LONG RAW",
    "DB_TYPE_NCHAR": "NCHAR",
    "DB_TYPE_NCLOB": "NCLOB",
    "DB_TYPE_NUMBER": "NUMBER",
    "DB_TYPE_NVARCHAR": "NVARCHAR2",
    "DB_TYPE_OBJECT": "OBJECT",
    "DB_TYPE_RAW": "RAW",
    "DB_TYPE_ROWID": "ROWID",
    "DB_TYPE_TIMESTAMP": "TIMESTAMP",
    "DB_TYPE_TIMESTAMP_LTZ": "TIMESTAMP WITH LOCAL TZ",
    "DB_TYPE_TIMESTAMP_TZ": "TIMESTAMP WITH TZ",
    "DB_TYPE_VARCHAR": "VARCHAR2"
}

class OracleConnectionMethod(enum.Enum):
    HOST = 1
    TNS = 2
    CONNECTION_STRING = 3


@dataclass
class OracleAdapterCredentials(Credentials):
    """Collect Oracle credentials

    An OracleConnectionMethod is inferred from the combination
    of parameters profiled in the profile.
    """
    # Mandatory required arguments.
    user: Optional[str] = None
    password: Optional[str] = None

    # Specifying database is optional
    database: Optional[str] = None
    schema: Optional[str] = None

    # OracleConnectionMethod.TNS
    tns_name: Optional[str] = None

    # OracleConnectionMethod.HOST
    protocol: Optional[str] = None
    host: Optional[str] = None
    port: Optional[Union[str, int]] = None
    service: Optional[str] = None

    # OracleConnectionMethod.CONNECTION_STRING
    connection_string: Optional[str] = None

    # shardingkey and supershardingkey is a list
    shardingkey: Optional[List[str]] = field(default_factory=list)
    supershardingkey: Optional[List[str]] = field(default_factory=list)

    # Database Resident Connection Pooling (DRCP)
    cclass: Optional[str] = None
    purity: Optional[str] = None

    # Connection retry params
    retry_count: Optional[int] = 1
    retry_delay: Optional[int] = 3

    # Base URL for ADB-S OML REST API
    oml_cloud_service_url: Optional[str] = None

    # session info is stored in v$session for each dbt run
    session_info: Optional[Dict[str, str]] = field(default_factory=dict)


    _ALIASES = {
        'dbname': 'database',
        'pass': 'password',
    }

    @property
    def type(self):
        return 'oracle'

    @property
    def unique_field(self):
        return self.database or self.user

    def _connection_keys(self) -> Tuple[str]:
        """
        List of keys to display in the `dbt debug` output. Omit password.
        """
        return (
            'user', 'database', 'schema',
            'protocol', 'host', 'port', 'tns_name',
            'service', 'connection_string',
            'shardingkey', 'supershardingkey',
            'cclass', 'purity', 'retry_count',
            'retry_delay', 'oml_cloud_service_url',
            'session_info'
        )

    @classmethod
    def __pre_deserialize__(cls, data: Dict[Any, Any]) -> Dict[Any, Any]:
        # If database is not defined as adapter credentials
        data = super().__pre_deserialize__(data)
        if "database" not in data:
            data["database"] = None
        return data

    def connection_method(self) -> OracleConnectionMethod:
        """Return an OracleConnecitonMethod inferred from the configuration"""
        if self.connection_string:
            return OracleConnectionMethod.CONNECTION_STRING
        elif self.host:
            return OracleConnectionMethod.HOST
        else:
            return OracleConnectionMethod.TNS

    def get_dsn(self) -> str:
        """Create dsn for cx_Oracle for either any connection method

        See https://cx-oracle.readthedocs.io/en/latest/user_guide/connection_handling.html"""

        method = self.connection_method()
        if method == OracleConnectionMethod.TNS:
            return self.tns_name
        if method == OracleConnectionMethod.CONNECTION_STRING:
            return self.connection_string

        # Assume host connection method OracleConnectionMethod.HOST and necessary parameters are defined
        return f'{self.protocol}://{self.host}:{self.port}/{self.service}?retry_count={self.retry_count}&retry_delay={self.retry_delay}'


class OracleAdapterConnectionManager(SQLConnectionManager):
    TYPE = 'oracle'

    @staticmethod
    def get_session_info(credentials):
        default_action = "DBT RUN"
        default_client_identifier = f'dbt-oracle-client-{uuid.uuid4()}'
        default_client_info = "_".join([platform.node(), platform.machine()])
        default_module = f'dbt-{dbt_version}'
        return {
            "action": credentials.session_info.get("action", default_action),
            "client_identifier": credentials.session_info.get("client_identifier", default_client_identifier),
            "clientinfo": credentials.session_info.get("client_info", default_client_info),
            "module": credentials.session_info.get("module", default_module)
        }

    @classmethod
    def open(cls, connection):
        if connection.state == 'open':
            logger.debug('Connection is already open, skipping open.')
            return connection
        credentials = cls.get_credentials(connection.credentials)
        method = credentials.connection_method()
        dsn = credentials.get_dsn()

        logger.debug(f"Attempting to connect using Oracle method: '{method}' "
                     f"and dsn: '{dsn}'")

        conn_config = {
            'user': credentials.user,
            'password': credentials.password,
            'dsn': dsn
        }

        if oracledb.__name__ == "oracledb":
            conn_config['connection_id_prefix'] = f'dbt-oracle-{dbt_version}-'

        if credentials.shardingkey:
            conn_config['shardingkey'] = credentials.shardingkey

        if credentials.supershardingkey:
            conn_config['supershardingkey'] = credentials.supershardingkey

        if credentials.cclass:
            conn_config['cclass'] = credentials.cclass

        if credentials.purity:
            purity = credentials.purity.lower()
            if credentials.purity == 'new':
                conn_config['purity'] = oracledb.ATTR_PURITY_NEW
            elif purity == 'self':
                conn_config['purity'] = oracledb.ATTR_PURITY_SELF
            elif purity == 'default':
                conn_config['purity'] = oracledb.ATTR_PURITY_DEFAULT

        if SQLNET_ORA_CONFIG is not None:
            conn_config.update(SQLNET_ORA_CONFIG)

        try:
            handle = oracledb.connect(**conn_config)
            # client_identifier and module are saved in corresponding columns in v$session
            session_info = cls.get_session_info(credentials=credentials)
            logger.debug(f"Session info :{json.dumps(session_info)}")
            for k, v in session_info.items():
                try:
                    setattr(handle, k, v)
                except AttributeError:
                    logger.warning(f"Python driver does not support setting {k}")

            connection.handle = handle
            connection.state = 'open'
        except oracledb.DatabaseError as e:
            logger.info(f"Got an error when attempting to open an Oracle "
                        f"connection: '{e}'")
            connection.handle = None
            connection.state = 'fail'

            raise FailedToConnectError(str(e))

        return connection

    @classmethod
    def cancel(cls, connection):
        connection_name = connection.name
        oracle_connection = connection.handle

        logger.info("Cancelling query '{}' ".format(connection_name))

        try:
            oracledb.Connection.close(oracle_connection)
        except Exception as e:
            logger.error('Error closing connection for cancel request')
            raise Exception(str(e))

        logger.info("Canceled query '{}'".format(connection_name))

    @classmethod
    def get_status(cls, cursor):
        # Do oracle cx has something for this? could not find it
        return 'OK'

    @classmethod
    def get_response(cls, cursor):
        # number of rows fetched for a SELECT statement or
        # have been affected by INSERT, UPDATE, DELETE and MERGE statements
        return AdapterResponse(rows_affected=cursor.rowcount,
                               _message='OK')

    @contextmanager
    def exception_handler(self, sql):
        try:
            yield

        except oracledb.DatabaseError as e:
            logger.info('Oracle error: {}'.format(str(e)))

            try:
                # attempt to release the connection
                self.release()
            except oracledb.Error:
                logger.info("Failed to release connection!")
                pass

            raise dbt_common.exceptions.DbtDatabaseError(str(e).strip()) from e

        except Exception as e:
            logger.info("Rolling back transaction.")
            self.release()
            if isinstance(e, dbt_common.exceptions.DbtRuntimeError):
                # during a sql query, an internal to dbt exception was raised.
                # this sounds a lot like a signal handler and probably has
                # useful information, so raise it without modification.
                raise e

            raise dbt_common.exceptions.DbtRuntimeError(str(e)) from e

    @classmethod
    def get_credentials(cls, credentials):
        return credentials

    def add_query(
        self,
        sql: str,
        auto_begin: bool = True,
        bindings: Optional[Any] = {},
        abridge_sql_log: bool = False
    ) -> Tuple[oracledb.Connection, Any]:
        connection = self.get_thread_connection()
        if auto_begin and connection.transaction_open is False:
            self.begin()

        fire_event(
            ConnectionUsed(
                conn_type=self.TYPE,
                conn_name=cast_to_str(connection.name),
                node_info=get_node_info(),
            )
        )

        with self.exception_handler(sql):
            if abridge_sql_log:
                log_sql = '{}...'.format(sql[:512])
            else:
                log_sql = sql

            fire_event(
                SQLQuery(
                    conn_name=cast_to_str(connection.name), sql=log_sql, node_info=get_node_info()
                )
            )

            pre = time.time()
            cursor = connection.handle.cursor()
            cursor.execute(sql, bindings)
            fire_event(
                SQLQueryStatus(
                    status=str(self.get_response(cursor)),
                    elapsed=round((time.time() - pre)),
                    node_info=get_node_info(),
                )
            )
            return connection, cursor

    def add_begin_query(self):
        connection = self.get_thread_connection()
        cursor = connection.handle.cursor
        return connection, cursor

    @classmethod
    def data_type_code_to_name(cls, type_code) -> str:
        return DATATYPES[type_code.name]

    def commit(self):
        connection = self.get_thread_connection()
        fire_event(SQLCommit(conn_name=connection.name, node_info=get_node_info()))
        self.add_commit_query()
        connection.transaction_open = False
        return connection



"""
Copyright (c) 2022, Oracle and/or its affiliates.
Copyright (c) 2020, Vitor Avancini

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
"""

"""The setup script."""
import sys
from setuptools import setup

try:
    from setuptools import find_namespace_packages
except ImportError:
    # the user has a downlevel version of setuptools.
    print("Error: dbt requires setuptools v40.1.0 or higher.")
    print('Please upgrade setuptools with "pip install --upgrade setuptools" ' "and try again")
    sys.exit(1)


# lockstep with dbt-core which requires Python > 3.8
if sys.version_info < (3, 8):
    print("Error: dbt-oracle does not support this version of Python.")
    print("Please upgrade to Python 3.8 or higher.")
    sys.exit(1)


with open('README.md') as readme_file:
    readme = readme_file.read()


requirements = [
        "dbt-common>=1.1.0,<2.0",
        "dbt-adapters>=1.2.1,<2.0",
        "dbt-core~=1.8,<1.9",
        "oracledb==2.3.0"
]

test_requirements = [
    "dbt-tests-adapter~=1.8,<1.9",
    "pytest"
]

project_urls = {
    'Documentation': 'https://docs.getdbt.com/reference/warehouse-profiles/oracle-profile',
    'Source': 'https://github.com/oracle/dbt-oracle',
    'Bug Tracker': 'https://github.com/oracle/dbt-oracle/issues',
    'CI': 'https://github.com/oracle/dbt-oracle/actions',
    "Release Notes": "https://github.com/oracle/dbt-oracle/releases"
}

url = 'https://github.com/oracle/dbt-oracle'

VERSION = '1.8.1'
setup(
    author="Oracle",
    python_requires='>=3.8',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12'
    ],
    description="dbt (data build tool) adapter for Oracle Autonomous Database",
    install_requires=requirements,
    license="Apache Software License 2.0",
    long_description=readme,
    long_description_content_type='text/markdown',
    include_package_data=True,
    keywords='Oracle dbt',
    name='dbt-oracle',
    packages=find_namespace_packages(include=["dbt", "dbt.*"]),
    test_suite='tests',
    tests_require=test_requirements,
    scripts=['bin/create-pem-from-p12'],
    url=url,
    project_urls=project_urls,
    version=VERSION,
    zip_safe=False
)



"""
Copyright (c) 2023, Oracle and/or its affiliates.

  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

     https://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
"""

from dataclasses import dataclass
from typing import Any, Dict, Optional

import agate
from dbt.adapters.base.relation import Policy
from dbt.adapters.relation_configs import (
    RelationConfigBase,
    RelationResults,
)
from dbt.contracts.graph.nodes import ModelNode
from dbt.adapters.contracts.relation import ComponentName

from dbt.adapters.oracle.relation_configs.policies import (
    OracleQuotePolicy,
    OracleIncludePolicy,
)


@dataclass(frozen=True, eq=True, unsafe_hash=True)
class OracleRelationConfigBase(RelationConfigBase):
    """
    This base class implements a few boilerplate methods and provides some light structure for Oracle relations.
    """

    @classmethod
    def include_policy(cls) -> Policy:
        return OracleIncludePolicy()

    @classmethod
    def quote_policy(cls) -> Policy:
        return OracleQuotePolicy()

    @classmethod
    def from_model_node(cls, model_node: ModelNode):
        relation_config = cls.parse_model_node(model_node)
        relation = cls.from_dict(relation_config)
        return relation

    @classmethod
    def parse_model_node(cls, model_node: ModelNode) -> Dict[str, Any]:
        raise NotImplementedError(
            "`parse_model_node()` needs to be implemented on this RelationConfigBase instance"
        )

    @classmethod
    def from_relation_results(cls, relation_results: RelationResults):
        relation_config = cls.parse_relation_results(relation_results)
        relation = cls.from_dict(relation_config)
        return relation

    @classmethod
    def parse_relation_results(cls, relation_results: RelationResults) -> Dict[str, Any]:
        raise NotImplementedError(
            "`parse_relation_results()` needs to be implemented on this RelationConfigBase instance"
        )

    @classmethod
    def _render_part(cls, component: ComponentName, value: Optional[str]) -> Optional[str]:
        if cls.include_policy().get_part(component) and value:
            if cls.quote_policy().get_part(component):
                return f'"{value}"'
            return value.lower()
        return None

    @classmethod
    def _get_first_row(cls, results: agate.Table) -> agate.Row:
        try:
            return results.rows[0]
        except IndexError:
            return agate.Row(values=set())
